/*BEGIN_LEGAL 
Intel Open Source License 

Copyright (c) 2002-2005 Intel Corporation 
All rights reserved. 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.  Redistributions
in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.  Neither the name of
the Intel Corporation nor the names of its contributors may be used to
endorse or promote products derived from this software without
specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
END_LEGAL */

//mcfTRaptor.cpp
//Multithreaded branch, call, and return instruction tracing 
//using TRaptor branch prediction
//
//Bugs and criticisms can be directed to
//myers.albert@gmail.com

#include <unistd.h>
#include <iomanip>
#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <deque>
#include <bitset>
#include <sys/time.h>
#include <sys/stat.h>
#include "pin.H"
#include "instlib.H"
#include "mcfTRaptorAssist.h"
#include "mcfTRaptor.h"
#include "mcfTRaptorPrivateAnalysis.h"
#include "mcfTRaptorSharedAnalysis.h"

//Can be used to filter which parts of the binary to instrument
INSTLIB::FILTER filter;

VOID ASCIITrace(TRACE trace, VOID *v)
{
   if( !filter.SelectTrace(trace) )
      return;

   for(BBL bbl = TRACE_BblHead(trace); BBL_Valid(bbl); bbl = BBL_Next(bbl) )
   {
      for(INS ins = BBL_InsHead(bbl); INS_Valid(ins); ins = INS_Next(ins) )
      {
         INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)SetFastForwardAndLength, IARG_THREAD_ID, IARG_END);

         //From the pin manual
         //HasFallThrough is TRUE for instructions which don't change the control flow (most instructions), 
         //or for conditional branches (which might change the control flow, but might not),
         //HasFallThrough is FALSE for unconditional branches and calls 
         //(where the next instruction to be executed is always explicitly specified).
         if( TRaptorShare.Value() )
         {
            //is Conditional and Direct
            //check outcome for misprediction
            if(INS_IsDirectBranchOrCall(ins) && INS_HasFallThrough(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_DirectConditional_ASCII,
                              //Args: Thread ID, Instruction Address, Taken
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TAKEN, IARG_END);

            // is return, should use RAS
            else if(INS_IsRet(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_Ret_ASCII,
                             //Args: Thread ID, Instruction Address, Target Address 
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_END);

            // use iBTB and push return address on RAS
            else if(INS_IsIndirectBranchOrCall(ins) && INS_IsCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_IndirectCall_ASCII,
                              //Args: Thread ID, Instruction Address, Target Address, return address
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_ADDRINT, INS_NextAddress(ins), IARG_END);

            // is other indirect branch, should use iBTB
            else if(INS_IsIndirectBranchOrCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_OtherUnconditionalIndirect_ASCII,
                              //Args: Thread ID, Instruction Address, Target Address 
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_END);

            //if direct call, we still need to push on the RAS 
            else if(INS_IsCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_Direct_Call, IARG_THREAD_ID, IARG_INST_PTR, IARG_ADDRINT, INS_NextAddress(ins), IARG_END);

            //increment iCnt for every other instruction
            else
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_iCnt_Increment, IARG_THREAD_ID, IARG_END);
         }
         else
         {
            //is Conditional and Direct
            //check outcome for misprediction
            if(INS_IsDirectBranchOrCall(ins) && INS_HasFallThrough(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_DirectConditional_ASCII,
                              //Args: Thread ID, Instruction Address, Taken
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TAKEN, IARG_END);

            // is return, should use RAS
            else if(INS_IsRet(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_Ret_ASCII,
                              //Args: Thread ID, Instruction Address, Target Address 
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_END);

            // use iBTB and push return address on RAS
            else if(INS_IsIndirectBranchOrCall(ins) && INS_IsCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_IndirectCall_ASCII,
                              //Args: Thread ID, Instruction Address, Target Address, return address
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_ADDRINT, INS_NextAddress(ins), IARG_END);

            // is other indirect branch, should use iBTB
            else if(INS_IsIndirectBranchOrCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_OtherUnconditionalIndirect_ASCII,
                              //Args: Thread ID, Instruction Address, Target Address 
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_END);

            //if direct call, we still need to push on the RAS 
            else if(INS_IsCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_Direct_Call, IARG_THREAD_ID, IARG_INST_PTR, IARG_ADDRINT, INS_NextAddress(ins), IARG_END);

            //increment iCnt for every other instruction
            else
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_iCnt_Increment, IARG_THREAD_ID, IARG_END);
         }
      }            
   }   
}

//Instrument ASCII with annotated disassembly
VOID DisTrace(TRACE trace, VOID *v)
{
   if( !filter.SelectTrace(trace) )
      return;

   for(BBL bbl = TRACE_BblHead(trace); BBL_Valid(bbl); bbl = BBL_Next(bbl) )
   {
      for(INS ins = BBL_InsHead(bbl); INS_Valid(ins); ins = INS_Next(ins) )
      {
         INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)SetFastForwardAndLength, IARG_THREAD_ID, IARG_END);

         if( TRaptorShare.Value())
         {   
            //is Conditional and Direct
            //check outcome for misprediction
            if(INS_IsDirectBranchOrCall(ins) && INS_HasFallThrough(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_DirectConditional_Dis,
                              //Args: Thread ID, Instruction Address, Taken
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TAKEN, IARG_END);

            // is return, should use RAS
            else if(INS_IsRet(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_Ret_Dis,
                              //Args: Thread ID, Instruction Address, Target Address 
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_END);

            // use iBTB and push return address on RAS
            else if(INS_IsIndirectBranchOrCall(ins) && INS_IsCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_IndirectCall_Dis,
                              //Args: Thread ID, Instruction Address, Target Address, return address
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_ADDRINT, INS_NextAddress(ins), IARG_END);

            // is other indirect branch, should use iBTB
            else if(INS_IsIndirectBranchOrCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_OtherUnconditionalIndirect_Dis,
                              //Args: Thread ID, Instruction Address, Target Address 
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_END);

            //if direct call, we still need to push on the RAS 
            else if(INS_IsCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_Direct_Call, IARG_THREAD_ID, IARG_INST_PTR, IARG_ADDRINT, INS_NextAddress(ins), IARG_END);

            //increment iCnt for every other instruction
            else
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_iCnt_Increment, IARG_THREAD_ID, IARG_END);

         }
         else
         {
            //is Conditional and Direct
            //check outcome for misprediction
            if(INS_IsDirectBranchOrCall(ins) && INS_HasFallThrough(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_DirectConditional_Dis,
                              //Args: Thread ID, Instruction Address, Taken
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TAKEN, IARG_END);

            // is return, should use RAS
            else if(INS_IsRet(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_Ret_Dis,
                              //Args: Thread ID, Instruction Address, Target Address 
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_END);

            // use iBTB and push return address on RAS
            else if(INS_IsIndirectBranchOrCall(ins) && INS_IsCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_IndirectCall_Dis,
                              //Args: Thread ID, Instruction Address, Target Address, return address
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_ADDRINT, INS_NextAddress(ins), IARG_END);

            // is other indirect branch, should use iBTB
            else if(INS_IsIndirectBranchOrCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_OtherUnconditionalIndirect_Dis,
                              //Args: Thread ID, Instruction Address, Target Address 
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_END);

            //if direct call, we still need to push on the RAS 
            else if(INS_IsCall(ins))                                     
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_Direct_Call, IARG_THREAD_ID, IARG_INST_PTR, IARG_ADDRINT, INS_NextAddress(ins), IARG_END);

            //increment iCnt for every other instruction
            else
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_iCnt_Increment, IARG_THREAD_ID, IARG_END); 
         }
      }
   }
}
//binary descriptors
VOID BinTrace(TRACE trace, VOID *v)
{

   if( !filter.SelectTrace(trace) )
      return;

   for(BBL bbl = TRACE_BblHead(trace); BBL_Valid(bbl); bbl = BBL_Next(bbl) )
   {
      for(INS ins = BBL_InsHead(bbl); INS_Valid(ins); ins = INS_Next(ins) )
      {
         INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)SetFastForwardAndLength, IARG_THREAD_ID, IARG_END);

         if(TRaptorShare.Value())
         {        
            //is Conditional and Direct
            //check outcome for misprediction
            if(INS_IsDirectBranchOrCall(ins) && INS_HasFallThrough(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_DirectConditional_Bin,
                              //Args: Thread ID, Instruction Address, Taken
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TAKEN, IARG_END);

            // is return, should use RAS
            else if(INS_IsRet(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_Ret_Bin,
                              //Args: Thread ID, Instruction Address, Target Address 
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_END);

            // use iBTB and push return address on RAS
            else if(INS_IsIndirectBranchOrCall(ins) && INS_IsCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_IndirectCall_Bin,
                              //Args: Thread ID, Instruction Address, Target Address, return address
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_ADDRINT, INS_NextAddress(ins), IARG_END);

            // is other indirect branch, should use iBTB
            else if(INS_IsIndirectBranchOrCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_OtherUnconditionalIndirect_Bin,
                              //Args: Thread ID, Instruction Address, Target Address 
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_END);

            //if direct call, we still need to push on the RAS 
            else if(INS_IsCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_Direct_Call, IARG_THREAD_ID, IARG_INST_PTR, IARG_ADDRINT, INS_NextAddress(ins), IARG_END);

            //increment iCnt for every other instruction
            else
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Shared_iCnt_Increment, IARG_THREAD_ID, IARG_END);

         }
         else
         {
            //is Conditional and Direct
            //check outcome for misprediction
            if(INS_IsDirectBranchOrCall(ins) && INS_HasFallThrough(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_DirectConditional_Bin,
                              //Args: Thread ID, Instruction Address, Taken
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TAKEN, IARG_END);

            // is return, should use RAS
            else if(INS_IsRet(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_Ret_Bin,
                              //Args: Thread ID, Instruction Address, Target Address 
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_END);

            // use iBTB and push return address on RAS
            else if(INS_IsIndirectBranchOrCall(ins) && INS_IsCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_IndirectCall_Bin,
                              //Args: Thread ID, Instruction Address, Target Address, return address
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_ADDRINT, INS_NextAddress(ins), IARG_END);

            // is other indirect branch, should use iBTB
            else if(INS_IsIndirectBranchOrCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_OtherUnconditionalIndirect_Bin,
                              //Args: Thread ID, Instruction Address, Target Address 
                              IARG_THREAD_ID, IARG_INST_PTR, IARG_BRANCH_TARGET_ADDR, IARG_END);

            //if direct call, we still need to push on the RAS 
            else if(INS_IsCall(ins))
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_Direct_Call, IARG_THREAD_ID, IARG_INST_PTR, IARG_ADDRINT, INS_NextAddress(ins), IARG_END);

            //increment iCnt for every other instruction
            else
               INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)Private_iCnt_Increment, IARG_THREAD_ID, IARG_END);

         }
      }
   }
}

VOID Fini(INT32 code, VOID *v)
{
   gettimeofday(&t2, NULL);
   //write remaining descriptors to file
   if(ASCII.Value())
   {
      if(usingCompression)
         WriteASCIIDescriptorTableToCompressor(PIN_GetTid()+1);
      else
         WriteASCIIDescriptorTableToFile(PIN_GetTid());
   }
   else
   {
      if(usingCompression)
         WriteBinaryDescriptorTableToCompressor(PIN_GetTid());
      else
         WriteBinaryDescriptorTableToFile(PIN_GetTid());
   }
   //close 
   closeOutput();
   if( TRaptorShare.Value() )
      delete SharedTRaptor;
   PrintStatistics();
}
//Called if Pin detaches from application
VOID DetachCallback(VOID *args)
{
   gettimeofday(&t2, NULL);
   //write remaining descriptors to file
   if(ASCII.Value())
   {
      if(usingCompression)
         WriteASCIIDescriptorTableToCompressor(PIN_GetTid()+1);
      else
         WriteASCIIDescriptorTableToFile(PIN_GetTid());
   }
   else
   {
      if(usingCompression)
         WriteBinaryDescriptorTableToCompressor(PIN_GetTid());
      else
         WriteBinaryDescriptorTableToFile(PIN_GetTid());
   }

   //close file/pipe
   closeOutput();
   if( TRaptorShare.Value() )
      delete SharedTRaptor;

   std::cout << "mcfTRaptor: Detaching..." << std::endl;
   PrintStatistics();
}

//Setup private TRaptor structures on thread start
VOID PrivateTRaptorThreadStart(THREADID threadid, CONTEXT *ctxt, INT32 flags, VOID *v)
{
   GetLock(&cout_lock, threadid+1);   
   //create TRaptor structures for each thread;
   TRaptor  *traptor = new TRaptor;
   PIN_SetThreadData(tls_key, traptor, threadid);
   numThreads++;
   std::cout << "mcfTRaptor: Private TRaptor thread begin " << static_cast<UINT16>(traptor->tid) << " " << PIN_GetTid() << std::endl;
   ReleaseLock(&cout_lock);
}
VOID SharedTRaptorThreadStart(THREADID threadid, CONTEXT *ctxt, INT32 flags, VOID *v)
{
   GetLock(&cout_lock, threadid+1);
   localCounters *counters = new localCounters(numThreads);
   PIN_SetThreadData(tls_key, counters, threadid);
   std::cout << "mcfTRaptor: Shared TRpator thread begin " << static_cast<UINT16>(counters->tid) << " " << PIN_GetTid() << std::endl;
   numThreads++;
   ReleaseLock(&cout_lock);
}

//delete thread local storage
VOID ThreadFini(THREADID threadid, const CONTEXT *ctxt, INT32 flags, VOID *v)
{
   if( !TRaptorShare.Value() )
      delete static_cast<TRaptor*>(PIN_GetThreadData(tls_key, threadid));
   else
      delete static_cast<localCounters*>(PIN_GetThreadData(tls_key, threadid));
}

INT32 Usage()
{
    cerr << KnobOutputFile.StringKnobSummary() << endl;
    return -1;
}

int main(int argc, char * argv[])
{
   // Initialize pin
   PIN_InitSymbols();
   if (PIN_Init(argc, argv)) return Usage();
  
   if(KnobOutputFile.Value() == "mcfTRaptor.out")
   {
      //Create time stamp for output file
      //Used so user won't overwrite previously generated files 
      time_t t = time(NULL);
      stamp = *localtime(&t);
      string temp;
      temp = std::to_string(static_cast<long long>(stamp.tm_year + 1900)) + "_" + std::to_string(static_cast<long long>(stamp.tm_mon + 1))+ "_"
             + std::to_string(static_cast<long long>(stamp.tm_mday)) + "_" + std::to_string(static_cast<long long>(stamp.tm_hour)) + "."
             + std::to_string(static_cast<long long>(stamp.tm_min)) + "." + std::to_string(static_cast<long long>(stamp.tm_sec));
      //Concatenate file name and time stamp;
      fileName = KnobOutputFile.Value() + temp;
   }
   else
      fileName = KnobOutputFile.Value();

   //Initialize locks
   InitLock(&table_lock);
   InitLock(&decoder_lock);
   InitLock(&cout_lock);
   InitLock(&count_lock);
   InitLock(&shared_lock);
   
   //The number of instructions to trace 
   if(Length.Value()==0)
      noLength = TRUE;
   else
      noLength = FALSE;

 
   //If writing descriptors to binary file
   if(!ASCII.Value())
   {
      if(AnnotateDisassembly.Value())
      {
         std::cout << "mcfTRaptor: Can only annotate disassembly if writing to ascii file" << std::endl;
         std::exit(-1);
      } 
      std::cout << "mcfTRaptor: Writing to binary file: " << fileName + ".bin" << std::endl;
      std::cout << "mcfTRaptor: mcfTRaptor descriptors -" << std::endl;
      std::cout << "\tMispredicted outcome for direct conditional branches: Descriptor Type (1 Byte), Thread ID (1 Byte), bCnt (4 Bytes)" << std::endl;
      std::cout << "\tMispredicted target for indirect unconditional branches: Descriptor Type (1 Byte), Thread ID (1 Byte), bCnt (4 Bytes), T (1 Bytes), Target Address (8 Bytes)" << std::endl;
      std::cout << "\tExceptions: Descriptor Type (1 Byte), Thread ID (1 Byte), bCnt (4 Bytes), iCnt (4 Bytes), Target Address (8 Bytes)" << std::endl;
      TRACE_AddInstrumentFunction(BinTrace, NULL);
   }
   //If writing descriptors to ASCII file
   else
   {
      std::cout << "mcfTRaptor: Writing to ASCII file: " << fileName + ".txt" << std::endl;
      std::cout << "mcfTRaptor: mcfTRaptor descriptors -" << std::endl;
      std::cout << "\tMispredicted outcomes for direct conditional branches: Thread ID, bCnt" << std::endl;
      std::cout << "\tMispredicted targets for indirect unconditional branches: Thread ID, bCnt, T, Target Address" << std::endl;
      std::cout << "\tExceptions: Thread ID, bCnt, iCnt, Target Address" << std::endl;

      if( AnnotateDisassembly.Value() )
      {
         xed_tables_init();
         TRACE_AddInstrumentFunction(DisTrace, NULL);
      }
      else
         TRACE_AddInstrumentFunction(ASCIITrace, NULL);
   } 
 
   //Open output file/pipe
   openOutput();
   //megabytes to bytes, go with power of ten for disk
   //will detach if larger than 
   fileLimit = FileLimit.Value() * 1000000;

   //use shared TRaptor 
   if( TRaptorShare.Value() )
   { 
      //allocate shared TRaptor
      SharedTRaptor = new TRaptor;
      PIN_AddThreadStartFunction(SharedTRaptorThreadStart, NULL);
   }
   //use private structures
   else
      PIN_AddThreadStartFunction(PrivateTRaptorThreadStart, NULL);
   

   //Register a detach function to be called if the pin exits early
   PIN_AddDetachFunction(DetachCallback, NULL);

   //Add function to delete TLS on thread completion
   PIN_AddThreadFiniFunction(ThreadFini, NULL);
   //Register Fini to be called when the application exits
   PIN_AddFiniFunction(Fini, NULL);

   //Procedure called for exceptions
   if( TRaptorShare.Value() )
      PIN_AddContextChangeFunction(Shared_Sig, 0);
   else
      PIN_AddContextChangeFunction(Private_Sig, 0);

   //Enable routine filter
   filter.Activate();

   // Start the program, never returns
   gettimeofday(&t1, NULL);
   PIN_StartProgram();

   return 0;
}
      

