/*BEGIN_LEGAL 
Intel Open Source License 

Copyright (c) 2002-2005 Intel Corporation 
All rights reserved. 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.  Redistributions
in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.  Neither the name of
the Intel Corporation nor the names of its contributors may be used to
endorse or promote products derived from this software without
specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
END_LEGAL */

//mcfTRaptorAssist.h
//Contains globals and utility procedures

// Bugs and criticisms can be directed to
// myers.albert@gmail.com


//Address width
//extended
#if defined(TARGET_IA32E)
   #define ADDRWID 16
//32 bit
#else
   #define ADDRWID 8
#endif

//KNOB types used to parse tool parameters
KNOB<std::string> KnobOutputFile(KNOB_MODE_WRITEONCE, "pintool", 
                    "o",   "mcfTRaptor.out",    "specify trace output file name");
KNOB<BOOL> ASCII(KNOB_MODE_WRITEONCE, "pintool", 
                    "a",      "0",   "use ASCII output file instead of binary");
KNOB<UINT64> Skip(KNOB_MODE_WRITEONCE,"pintool", 
                    "s",      "0",   "Begin emitting branch descriptors after executing a specified number of instructions ");
KNOB<UINT64> Length(KNOB_MODE_WRITEONCE, "pintool", 
                    "l",      "0",   "Number of instructions to profile (default is no limit)");
KNOB<std::string> Compression(KNOB_MODE_WRITEONCE, "pintool",
                    "c",      "0", "Compress trace. Supports bzip2, pbzip2, gzip, and pigz");
KNOB<BOOL> AnnotateDisassembly(KNOB_MODE_WRITEONCE, "pintool", 
                    "d",      "0", "Annotate descriptors with disassembly (only works when output is ASCII)");
KNOB<UINT64> FileLimit(KNOB_MODE_WRITEONCE, "pintool", 
                    "f",  "50000", "Output file size limit in MB. Tracing will end after reaching this limit. Default is 50000 MB");
//TRaptor specific KNOBs
KNOB<BOOL> TRaptorShare(KNOB_MODE_WRITEONCE, "pintool", 
                    "TRaptorShare",   "0", "Share the same T-Raptor structure among each thread");
KNOB<UINT32> GSHAREsize(KNOB_MODE_WRITEONCE, "pintool",
                    "GShare",       "256", "GShare outcome predictor size for T-Raptor trace module (must be 0, 256, 512, or 1024, 2048, 4096)");
KNOB<UINT32> RASsize(KNOB_MODE_WRITEONCE, "pintool", 
                       "RAS",         "8", "Size of return address stack for T-Raptor (must be 0, 8, 16, or 32 entries)");
KNOB<UINT32> IBTBsize(KNOB_MODE_WRITEONCE, "pintool", 
                      "iBTB",        "64", "Size of 2-way set associative indirect branch target buffer for T-Raptor (must be 0, 16, 32, or 64 entries)");

//key for private TRaptor structures
TLS_KEY tls_key;

//Lock globals
//protects descriptor table
PIN_LOCK table_lock;
//protects xed decoder
PIN_LOCK decoder_lock;
//protects streams
PIN_LOCK cout_lock;
//protects statistics globals
PIN_LOCK count_lock;
//protects shared TRaptor analysis functions
PIN_LOCK shared_lock;

//Globals for file
struct tm stamp;
std::string fileName;
std::string statFileName;
//Used to find pin tool overhead time.
timeval t1, t2;
//ofstream for raw files
std::ofstream OutFile;
//FILE for piping to compression utility
FILE *outPipe;

UINT32 numThreads;

//output file size limit;
UINT64 fileCount;
UINT64 fileLimit;
//number of instructions to skip;
UINT64  fastforwardLength; 
//number of instructions to profile
UINT64 traceLength;
//if a length isn't specififed
BOOL noLength;
//if using compression
BOOL usingCompression;

//Total number of branches
//automatically initialized to zero
UINT64 numTotalBranches;
//records number of branches for each type
//global arrays are initialized to zero by default
UINT64 numBranches[3];
//only record prediction/mispredictions for conditional direct
//and unconditional indirect
UINT64 numMiss[2];
UINT64 numCorrect[2];

//A generic descriptor is pushed onto a deque for each misprediction
//When writing to file these descriptors are filtered down
//to the proper TRaptor descriptor
typedef enum : UINT8
{
   //mis-predicted outcome
   Outcome = 0,
   //mis-predicted targer
   Target = 1,
   //Exceptions
   Exception = 2,
} TRaptorDescriptorTypes;

//struct for pushing on Bin deque
typedef struct{
   TRaptorDescriptorTypes type;   
   //logical thread id
   UINT8 tid;
   UINT32 bCnt;
   UINT32 iCnt;
   ADDRINT targetAddress; // address of branch target
} BinaryDescriptorTableEntry;
//size of BinaryDescriptorTableEntries
const UINT8 GenericDescriptorSize = sizeof(UINT8)*2 + sizeof(UINT32)*2 + sizeof(ADDRINT);
const UINT8 OutcomeDescriptorSize = sizeof(UINT8) + sizeof(UINT32);
const UINT8 TargetDescriptorSize = sizeof(UINT8)*2 + sizeof(UINT32) + sizeof(ADDRINT);
const UINT8 ExceptionDescriptorSize = sizeof(UINT8) + sizeof(UINT32)*2 + sizeof(ADDRINT);

//Deques used to hold descriptors, which will be written to file by PIN thread
std::deque<BinaryDescriptorTableEntry> binDescriptorTable;
std::deque<std::string> ASCIIDescriptorTable;

//Since T-Raptor records different message types for outcome and target 
//mispredictions we need to check the type when writing to file
//write binary descriptor table to file
inline VOID WriteBinaryDescriptorTableToFile(const THREADID threadid)
{
   GetLock(&table_lock, threadid+1);
   while( !binDescriptorTable.empty() ) 
   {
      BinaryDescriptorTableEntry temp = binDescriptorTable.front();
      //DirCB
      if(temp.type == Outcome)
      {
         OutFile.write((char *)&temp.tid, sizeof(UINT8));
         OutFile.write((char *)&temp.bCnt, sizeof(UINT32));
      }
      //IndUB
      else if(temp.type == Target)
      {
         OutFile.write((char *)&temp.tid, sizeof(UINT8));
         OutFile.write((char *)&temp.bCnt, sizeof(UINT32));
         UINT8 taken = 255;        
         OutFile.write((char *)&taken, sizeof(UINT8));
         OutFile.write((char *)&temp.targetAddress,sizeof(ADDRINT)); 
      }
      //Except
      else
      {
         OutFile.write((char *)&temp.tid, sizeof(UINT8));
         UINT32 zero = 0;
         OutFile.write((char *)&zero, sizeof(UINT32));
         OutFile.write((char *)&temp.iCnt, sizeof(UINT32));
         OutFile.write((char *)&temp.targetAddress,sizeof(ADDRINT)); 
      }   
      binDescriptorTable.pop_front();
   }
   ReleaseLock(&table_lock);
}
//write binary descriptor table to compression program
inline VOID WriteBinaryDescriptorTableToCompressor(const THREADID threadid)
{
   GetLock(&table_lock, threadid+1);
   while( !binDescriptorTable.empty() ) 
   {
      BinaryDescriptorTableEntry temp = binDescriptorTable.front();
      fwrite(&temp.tid, sizeof(UINT8), 1,outPipe);
      //DirCB
      if(temp.type == Outcome)
         fwrite(&temp.bCnt, sizeof(UINT32), 1, outPipe);
      //IndUB
      else if(temp.type == Target)
      {
         fwrite(&temp.bCnt, sizeof(UINT32), 1, outPipe);
         UINT8 taken = 255; 
         fwrite(&taken, sizeof(UINT8), 1, outPipe);
         fwrite(&temp.targetAddress, sizeof(ADDRINT), 1, outPipe);
      }  
      //Except
      else
      {
         UINT32 zero = 0;
         fwrite(&zero, sizeof(UINT32), 1, outPipe);
         fwrite(&temp.iCnt, sizeof(UINT32), 1, outPipe);
         fwrite(&temp.targetAddress, sizeof(ADDRINT), 1, outPipe);
      }
      binDescriptorTable.pop_front();
    }
    ReleaseLock(&table_lock);
}

//write ASCII descriptor table to file
inline VOID WriteASCIIDescriptorTableToFile(const THREADID threadid)
{
   GetLock(&table_lock, threadid+1);
   //write contents to disk
   while( !ASCIIDescriptorTable.empty() )
   {
      OutFile << ASCIIDescriptorTable.front();
      ASCIIDescriptorTable.pop_front();
   }
   ReleaseLock(&table_lock);
}
//write ASCII descriptor table to compression program
inline VOID WriteASCIIDescriptorTableToCompressor(const THREADID threadid)
{
   GetLock(&table_lock, threadid+1);
   //write contents to disk
   while( !ASCIIDescriptorTable.empty() )
   {
      fputs(ASCIIDescriptorTable.front().c_str(), outPipe);
      ASCIIDescriptorTable.pop_front();
   }
   ReleaseLock(&table_lock);
}


//Worker ASCII write thread when not using compression
VOID ThreadWriteASCII(VOID *arg)
{
   THREADID threadid = PIN_ThreadId();
   if(usingCompression)
   {
      while(1)
      {
         //if process is closing (entered fini()) kill thread
         if( PIN_IsProcessExiting() )
            PIN_ExitThread(1);

         WriteASCIIDescriptorTableToCompressor(threadid);
      }
   }
   else
   {
      while(1)
      {
         //if process is closing (entered fini()) kill thread
         if( PIN_IsProcessExiting() )
            PIN_ExitThread(1);

         WriteASCIIDescriptorTableToFile(threadid);
      }
   }
}

//Worker bin write thread when not using compression
VOID ThreadWriteBin(VOID *arg)
{
   THREADID threadid = PIN_ThreadId();
   if(usingCompression)
   {
      while(1)
      {
         //if process is closing (entered fini()) kill thread
         if( PIN_IsProcessExiting() )
            PIN_ExitThread(1);

         WriteBinaryDescriptorTableToCompressor(threadid);
      }
   }
   else
   {
      while(1)
      {
         //if process is closing (entered fini()) kill thread
         if( PIN_IsProcessExiting() )
            PIN_ExitThread(1);

         WriteBinaryDescriptorTableToFile(threadid);
      }
   }
}
//check file size and exit early if too large
inline VOID IncrementFileCount(UINT32 size)
{
   //increment counter
   fileCount += size;
   //if limit reached, exit
   if(fileCount >= fileLimit)
      PIN_Detach();
}
inline VOID PushASCIIDescriptor(const THREADID threadid, const string & desc)
{
      GetLock(&table_lock, threadid+1);
      ASCIIDescriptorTable.push_back(desc);
      IncrementFileCount(desc.size());
      ReleaseLock(&table_lock);
}
inline VOID PushBinDescriptor(const THREADID threadid, const BinaryDescriptorTableEntry & desc)
{
      GetLock(&table_lock, threadid+1);
      binDescriptorTable.push_back(desc);
      //TID
      if(desc.type == Outcome)
         IncrementFileCount(OutcomeDescriptorSize);
      else if(desc.type == Target)
         IncrementFileCount(TargetDescriptorSize);
      else
         IncrementFileCount(ExceptionDescriptorSize);

      ReleaseLock(&table_lock);
}

//Used to check fast forward and trace length in analysis functions
inline BOOL CanEmit(const THREADID threadid)
{
   GetLock(&count_lock, threadid+1);

   if( fastforwardLength < Skip.Value()  )
   {
      ReleaseLock(&count_lock);
      return false;
   }
   else
   {
      ReleaseLock(&count_lock);
      return true; 
   }
}

//ins contains the instruction string
inline VOID getAssemblyString(string & ins, const THREADID threadid, const ADDRINT address)
{
   GetLock(&decoder_lock, threadid+1);

   //Use Intel XED to decode the instruction based on its address
#if defined(TARGET_IA32E)
    static const xed_state_t dstate = {XED_MACHINE_MODE_LONG_64, XED_ADDRESS_WIDTH_64b};
#else
    static const xed_state_t dstate = { XED_MACHINE_MODE_LEGACY_32, XED_ADDRESS_WIDTH_32b};
#endif
    xed_decoded_inst_t xedd;
    xed_decoded_inst_zero_set_mode(&xedd,&dstate);

   unsigned int max_inst_len = 15;

    xed_error_enum_t xed_code = xed_decode(&xedd, reinterpret_cast<UINT8*>(address), max_inst_len);
    BOOL xed_ok = (xed_code == XED_ERROR_NONE);
    if (xed_ok) {
        char buf[2048];
        
        xed_decoded_inst_dump_intel_format(&xedd, buf, 2048, address);

        ins.assign(buf);
    }
    else
        ins = "XED failed to decode this instruction";

   ReleaseLock(&decoder_lock);
} 

inline void IncrementBranchStatistics(const TRaptorDescriptorTypes branchType, const BOOL correctPrediction)
{
   
   //atomics
   __sync_fetch_and_add(&numTotalBranches, 1);
   __sync_fetch_and_add(numBranches+branchType, 1);
  if(branchType != Exception)
  {
     if(correctPrediction)
        __sync_fetch_and_add(numCorrect+branchType, 1);
     else
        __sync_fetch_and_add(numMiss+branchType, 1);
  }
}

//return percentage
inline double Percentage(const UINT64 num, const UINT64 den)
{
   double ratio = (double)num/(double)den;
   return 100*ratio;
}
VOID PrintStatistics()
{
   std::ofstream StatFile;
   StatFile.open(fileName + ".Statistics");

   double elapsedTime;
   elapsedTime = (t2.tv_sec - t1.tv_sec) * 1000.0;
   elapsedTime += (t2.tv_usec - t1.tv_usec) / 1000.0;

   StatFile << "mcfTRaptor: Instrumentation Time " <<  elapsedTime << " ms" << std::endl;

   StatFile << "mcfTRaptor: Skipped " << fastforwardLength << " instructions" << std::endl;  

   StatFile << "mcfTRaptor: Recorded " << numTotalBranches << " direct conditional branches, indirect unconditional branches, and exceptions" << std::endl;
   StatFile.setf(std::ios::fixed);

   StatFile << numBranches[Outcome] << " conditional direct branches " << std::endl;
   StatFile << "\t" << numCorrect[Outcome] << " ( %" << std::setprecision(2)  << Percentage(numCorrect[Outcome],numBranches[Outcome]) << " ) outcomes predicted" << std::endl;
   StatFile << "\t" << numMiss[Outcome] << " ( %" << std::setprecision(2)  << Percentage(numMiss[Outcome],numBranches[Outcome]) << " ) outcomes mispredicted" << std::endl;

   StatFile << numBranches[Target] << " unconditional indirect branches " << std::endl;
   StatFile << "\t" << numCorrect[Target] << " ( %" << std::setprecision(2)  << Percentage(numCorrect[Target],numBranches[Target]) << " ) targets predicted" << std::endl;
   StatFile << "\t" << numMiss[Target] << " ( %" << std::setprecision(2)  << Percentage(numMiss[Target],numBranches[Target]) << " ) targets mispredicted" << std::endl;

   StatFile << numBranches[Exception] << " exceptions " << std::endl;

   //Albert 
   //File size
   //struct stat st;
   //stat(statFileName.c_str(), &st);
   //StatFile << static_cast<UINT32>(numThreads) << "," << st.st_size << std::endl;

   //StatFile << static_cast<UINT32>(numThreads) << "," << elapsedTime << std::endl;

}

VOID openOutput()
{

   //Check compression knob
   string compressor = Compression.Value();
   if(compressor != "0" && compressor != "bzip2" && compressor != "pbzip2" && compressor != "gzip" && compressor != "pigz")
   {
      std::cout << "mcfTRaptor: That compression program is not supported" << std::endl;
      std::exit(-1);
   }
   if(compressor != "0")
   {
      //const string temp = "/bin/" + compressor;
      if( access( string( "/bin/" + compressor ).c_str(), X_OK) == -1  && access( string( "/usr/bin/" + compressor ).c_str(), X_OK) == -1 ) 
      {
         std::cout << "mcfTRaptor: This system does not support " << compressor << ". Exiting...\n";
         std::exit(-1);
      }
      usingCompression = TRUE;
   }
   else
      usingCompression = FALSE;

   //attach extension
   string tempFileName;
   if(ASCII.Value() == FALSE)
      tempFileName = fileName + ".bin";
   else
      tempFileName = fileName + ".txt";
  

   if( usingCompression )
   {
      char temp[1000] = { 0 };
      if( compressor == "bzip2" )
      {
         tempFileName += ".bz2"; 
         sprintf(temp, "bzip2 > %s", tempFileName.c_str());
         outPipe = popen(temp, "w");
      }
      else if( compressor == "pbzip2" )
      {
         tempFileName += ".bz2"; 
         sprintf(temp, "pbzip2 > %s", tempFileName.c_str());
         outPipe = popen(temp, "w");
      }
      else if( compressor == "gzip" )
      {
         tempFileName += ".gz"; 
         sprintf(temp, "gzip > %s", tempFileName.c_str());
         outPipe = popen(temp, "w");
      }
      else
      {
         tempFileName += ".gz"; 
         sprintf(temp, "pigz > %s", tempFileName.c_str());
         outPipe = popen(temp, "w");
      }
   
      if( outPipe == NULL )
      { 
         std::cout << "mcfTRaptor: Error opening outPipe..." << std::endl;
         std::exit(-1);
      }
   }
   //if not using compression, write to raw file
   else
   {
      if( ASCII.Value() == FALSE )
         OutFile.open(tempFileName, ios::out | ios::binary);
      else
         OutFile.open(tempFileName);
   }
   statFileName = tempFileName;

   THREADID err;
   //Spawn write thread
   if( ASCII.Value() == FALSE )
      err = PIN_SpawnInternalThread(ThreadWriteBin, NULL, 0, NULL);
   else
      err = PIN_SpawnInternalThread(ThreadWriteASCII, NULL, 0, NULL);
 
   if( err==INVALID_THREADID )
      std::cout<<"Error creating internal Pin thread. Exiting.." << std::endl;
}
VOID closeOutput()
{
   if(usingCompression) 
      pclose(outPipe);
   else
      OutFile.close();


   ASCIIDescriptorTable.clear();
   binDescriptorTable.clear();
}
